#!/usr/bin/env python

# rdf2sql.py
# based on simpleSQL.py example for using pyparsing to parse the SQL queries
# simpleSQL.py does not handle UNION, INTERSECT, EXCEPT or nested SQLs
# It also does not handle mathematical or other expressions in the filter conditions (eg. WHERE AGE = 2 * 18)
#
# ankesh

import sys
from traceback import print_exc
from time import localtime, strftime
from getopt import getopt
from csv import reader
from httplib import HTTPConnection
from urllib import urlencode

from pyparsing import alphanums, alphas, Keyword, Optional, ParseException, Word
from simpleSQL import simpleSQL
from tmswap.llyn import RDFStore

qpa_namespace = "http://dig.csail.mit.edu/2009/IARPA-PIR/dry-run/qpa#"
default_namespace = "http://dig.csail.mit.edu/2009/IARPA-PIR/dry-run/qpa-query#"

#contains dictionary for tables (table, col1, col2, ....)
dbDictionary = "file:///home/ankesh/exam-harness/Policy-dryrun/db_dictionary.csv"

#expression for parsing 'TABLE.COLUMN' 
tableDotColumn = Word(alphas, alphanums) + Optional("." + Word(alphas, alphanums))
#expression for parsing 'user userid'
userName = Keyword("user", caseless=True) + Word(alphanums + "/:.-")

columnsOfTables = dict() # {'table': ['col1', 'col2'...], ...}
columnTypes = dict() # {'table' : {'col1':<number or string>, ...}, ...}

def formattedLocalTime() :
    	return strftime("%a, %d %b %Y %H:%M:%S + 0530", localtime())

def getPrefixDeclarations() :
	global qpa_namespace, default_namespace
	n3string="#$ Generated by code $\n"
	n3string+="#$ at %s $\n"%formattedLocalTime()
	n3string+="\n@prefix math: <http://www.w3.org/2000/10/swap/math#> ."
	n3string+="\n@prefix string: <http://www.w3.org/2000/10/swap/string#> ."
	n3string+="\n@prefix qpa: <%s> ."%qpa_namespace
	n3string+="\n@prefix : <%s> .\n"%default_namespace
	return n3string

def populateTableDescs() :
	global columnsOfTables, columnTypes
	if columnsOfTables <> {} : return
	if dbDictionary.endswith(".csv") :
		filename = dbDictionary.replace("file://","")
		lines = reader(open(filename), delimiter = ',', quotechar='|')
		for row in lines :
			table = ""
			columns = []
			for string in row :
				if table : columns.append(string.upper())
				else : table = string.upper()
			columnsOfTables[table] = columns
	else :
		#TODO - db dictionary described in rdf
		pass

def populateColumnTypes() :
	global columnTypes
	#TODO - 
	columnTypes = dict()

def setColumnType(table, column, type) :
	global columnTypes
	if not table : return
	if table not in columnTypes : columnTypes[table] = {column : type}
	elif column not in columnTypes : columnTypes[table].update({column : type })
	else : pass

def separateTableColumn(string, tables) :
        tokens = tableDotColumn.parseString(string)
        if len(tokens) == 3 :
                return (tokens[0], tokens[2])
        else :
                if len(tables) == 1 :
                        return (tables[0], tokens[0])
                else :
                        return (None, tokens[0]) #TODO

def getOperatorUri(table, column, operator, value) :
	global mapNumericOperators, mapStringOperators, columnTypes
	if table in columnTypes and column in columnTypes[table] :
		return mapNumericOperators[operator] if columnTypes[table][column] == 'number' else mapStringOperators[operator]	
	if isNumeric(value) : 
		setColumnType(table, column, 'number')
		return mapNumericOperators[operator]
	elif isQuotedString(value) : 
		setColumnType(table, column, 'string')
		return mapStringOperators[operator]
	elif not (operator in mapStringOperators) : 
		setColumnType(table, column, 'number')
		return mapNumericOperators[operator]
	#TODO :TABLE1.COLx (=|<>) TABLE2.COLy type filter condition. decide operator based on type(COLx) or type(COLy)
	#current approach is flawed
	else : 
		return mapStringOperators[operator]
	
def isNumeric(value) :
	try:
		float(value)
		return True
	except: 
		return False	

def isQuotedString(value) :
	return (value.startswith('"') and value.endswith('"'))

mapStringOperators = {"=":"string:matches", "!=":"string:notMatches", "<>":"string:notMatches"}
mapNumericOperators = {"=":"math:equalTo", "!=":"math:notEqualTo", "<>":"math:notEqualTo", ">":"math:greaterThan", "<":"math:lessThan", ">=":"math:notLessThan", "=>":"math:notLessThan", "<=":"math:notGreaterThan", "=<":"math:notGreaterThan"}

filterConditionIds = dict() #key = (lhs, op, rhs) where filter condition is 'lhs op rhs'

def getRDFofSQL(user, queryId, prevQueryId, query) :
    global columnsOfTables, filterConditionIds, mapStringOperators, mapNumericOperators
    ERROR = True
    try:
	presuf = ("<",">") if (user.startswith("http://") or user.endswith("file://")) else ('"','"')
	user = presuf[0] + user + presuf[1]
	n3query = '\n%s a qpa:Query ;\n'%queryId
	n3query += '	qpa:user %s ;\n'%user
	n3query += '	qpa:prevQuery %s ;\n'%prevQueryId if prevQueryId else ""
	#n3query += '	qpa:prevQuery %i ;\n'%queryId
        tokens = simpleSQL.parseString(query.replace('"', "'"))

	columns = tokens.columns if tokens.columns <> "*" \
		  else reduce(list.__add__, [columnsOfTables[table] for table in tokens.tables], [])
	
	n3fields = ""	
	for column in columns :
		(table, col) = separateTableColumn(column, tokens.tables)
		n3fields += "," if n3fields else "	qpa:field"
		if table :
			n3fields += '\n		[ qpa:table "%s" ;'%table
			n3fields += '\n		  qpa:column "%s" ]'%col
		else :
			n3fields += '\n		[ qpa:column "%s" ]'%col

	n3fields += " ;\n" if columns else ""
	n3fromtables = ""		
	for table in tokens.tables :
		n3fromtables += ', "%s"'%table if n3fromtables else '	qpa:fromTable "%s"'%table
	n3fromtables += " ;\n" if tokens.tables else ""
	
	if len(tokens.where)>1: 
		return ERROR, "Parse Exception %s\nQUERY = %s\nUnexpected expression in Where Clause : %s"%(formattedLocalTime(), query, tokens.where[1])
	
	n3whereconditions = ""
	conditionIdsList = ""
	conditionNum = 0
	n3aggrConditions = ""
	#num_of_filters = len(tokens.where[0]) / 2
	
	for condition in tokens.where[0] :
		if isinstance(condition, str) :
			COND = str(condition).upper()
			if COND == "WHERE" : pass
			elif COND == "AND" : pass
			elif COND == "OR" : pass #"OR" is, for now, treated as "AND"
			else : return ERROR, "Parse Exception %s\nQUERY = %s\nUnexpected Boolean Operator '%s'"%(formattedLocalTime(), query, str(condition).upper())
		else :
			if len(condition)<>3 or condition[1] == 'IN':  
				return ERROR, "Parse Exception %s\nQUERY = %s\nUnexpected Filter condition in where clause: %s"%(formattedLocalTime(), query, str(condition))
			conditionNum += 1
			attribute = condition[0]
			reloperator = condition[1]
			value = condition[2]
			(table, col) = separateTableColumn(attribute, tokens.tables)
			isNumber = isNumeric(value)
			reloperator = getOperatorUri(table, col, reloperator, value)
			key = ()
	
			n3condition = "	qpa:attribute\n"
			if table :
				n3condition += '		[ qpa:table "%s" ;\n'%table
				n3condition += '	  	  qpa:column "%s" ] ;\n'%col
				key += (table, col)
			else :
				n3condition += '		[ qpa:column "%s" ] ;\n'%col
				key += (col,)
			
			n3condition += "	qpa:relOperator %s ;\n"%reloperator
			key += (reloperator,)
			if isNumber :
				n3condition += "	qpa:value %s .\n"%str(value)
				key += (value,)
			else :
				value = value.replace("'", '"')		
				if isQuotedString(value) : #color = 'red' 
					n3condition += "	qpa:value %s .\n"%value
					key += (value,)
				else : #object.color = thing.color
					(table, col) = separateTableColumn(value, tokens.tables)
					n3condition += "	qpa:attribute-2\n"
					if table :
						n3condition += '		[ qpa:table "%s" ;\n'%table
						n3condition += '  		  qpa:column "%s" ] .\n'%col
						key += (table, col)
					else :
						n3condition += '		[ qpa:column "%s" ] .\n'%col	
						key += (col,)
			
			filterConditionId = ""
			if key in filterConditionIds: 
				filterConditionId = filterConditionIds[key]
			else :
				filterConditionId = "%s-filter-%i"%(queryId, conditionNum)
				filterConditionIds[key] = filterConditionId
				n3condition = '%s a qpa:FilterCondition ;\n'%filterConditionId + n3condition
				n3aggrConditions += "\n" + n3condition
		
			n3whereconditions += (", " if n3whereconditions else "	qpa:filterCondition ") + filterConditionId

                        if conditionIdsList :  conditionIdsList += " "
                        conditionIdsList += filterConditionId
			
			#n3condition = filterConditionId + " a qpa:FilterCondition ;\n" + n3condition
			#n3aggrConditions += "\n" + n3condition
	
	if n3whereconditions : 
		n3whereconditions += " ;\n"                  
	conditionIdsList = "(" + conditionIdsList + ")"
	n3query += n3fields + n3fromtables + n3whereconditions
	n3query += "	qpa:filterConditionsList %s .\n"%conditionIdsList
	n3query += n3aggrConditions
	
	return not ERROR, n3query
    except ParseException, err:
	return ERROR, "Parse Exception %s \n QUERY = %s\n"%(formattedLocalTime(), query)\
		+ " "*err.loc + "^\n%s\n%s"%(err.msg, str(err))
	
foafPrimaryTopics = dict()

def getFoafPrimaryTopic(foaf) :
	global foafPrimaryTopics
	ERROR = True
	if foaf in foafPrimaryTopics : return not ERROR, foafPrimaryTopics[foaf]
        store = RDFStore()
	try : 
        	f = store.load(foaf) #TODO - check that f contains triples from this foaf file only
        	topic = f.newSymbol("http://xmlns.com/foaf/0.1/primaryTopic")
        	statements = f.statementsMatching(pred = topic)
        	OBJ = 3
        	for s in statements :
			foafPrimaryTopics[foaf] = s[OBJ].uriref()
			return not ERROR, foafPrimaryTopics[foaf]
	except IOError, err :
		return ERROR, "Error in Getting the FOAF Primary Topic %s.\n FOAF FILE = %s.\n%s"%(formattedLocalTime(), foaf, err)

queryNum = -1

def rdf2sql(SQLQueries, user = None) :
	global queryNum
    	populateTableDescs()
	#populateColumnTypes()

	if not user : 
		user = "dummy-user"
	elif user.startswith("http://") or user.startswith("file://") :
		#when foaf file contains references to policy files are web retrievable. {foaf#me air:policy :Policy-name}
		# not used under current scheme
		# qpa:user value in policy must be the same as foaf:primaryTopicpolicy
		err, user = getFoafPrimaryTopic(user)
		if err :
			return True, "Unable to get value of foaf:primaryTopic"
	
	queries = []
	n3string = getPrefixDeclarations()
	
	for query in SQLQueries :
		n3string += "\n# %s"%query
		queryNum += 1
		queryId = ":query-%i"%queryNum
		prevQueryId = queries[len(queries)-1] if len(queries)>0 else None
		queries.append(queryId)
		err, rdfofsql = getRDFofSQL(user, queryId, prevQueryId, query)
		if not err : 
			n3string += "\n " + rdfofsql
		else :
			queries.pop()
			return True, rdfofsql
	return False, n3string
		
queries = \
	["SELECT AGE FROM PEOPLE",
	"SELECT * FROM PEOPLE",
	"SELECT AGE, FIRSTNAME FROM PEOPLE WHERE STATE = 'MA'",
	"SELECT AGE, FIRSTNAME FROM PEOPLE WHERE AGE >= 19",
	"SELECT AGE, FIRSTNAME FROM PEOPLE WHERE AGE = 14 AND STATE = 'MA' AND OCCUPATION = 'COMPANYFOUNDER'",
	"SELECT PEOPLE.FIRSTNAME, USDATA.STATUS FROM PEOPLE, USDATA WHERE PEOPLE.AGE = 14 AND PEOPLE.SSN = USDATA.SSN AND USDATA.CITIZEN = 'US'"]

def main(argv = None) :
	global queries, filterConditionIds
	rdf = rdf2sql(queries)
	print rdf[1]
	
        user = None
        SQLQueries = []
        done = False
        while not done :
                sys.stdout.write('\n\r>> ')
                sys.stdout.flush()
                command = ""
                while not command: 
			command = sys.stdin.readline().strip()
                COMMAND = command.upper()
                if COMMAND.startswith("SELECT") : 
			SQLQueries.append(command)
                elif COMMAND == "EXECUTE" :
                        rdf = rdf2sql(SQLQueries, user)
			if rdf[0] : print "ERROR: ", rdf[1]
			else : print rdf[1]
                        SQLQueries = []
                        filterConditionIds = dict()
                elif COMMAND.startswith("USER"): user = userName.parseString(command)[1]
                elif COMMAND == "DONE": done = True

if __name__ == "__main__" :
	main()
